"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import unittest
import os
import sys

from configparser import ConfigParser
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.parametertree import Parameter

from c2dataviewer.model import DataSource as DataReceiver
from c2dataviewer.view import StripToolConfigure
from c2dataviewer.striptool import StripToolWindow, WarningDialog, PvEditDialog

from c2dataviewer.control.striptool_config import StriptoolConfig
from c2dataviewer.control.striptool_controller import StripToolController
from c2dataviewer.control.pvconfig import PvConfig

os.environ["QT_QPA_PLATFORM"] = "offscreen"

class TestStriptoolController(unittest.TestCase):
    DEFAULT_CFG = """
[STRIPTOOL]
"""
    def setUp(self):
        # Create ImageWindow and get the imageWidget instance
        self.window = StripToolWindow()

        # Build parameter tree
        cf = ConfigParser()
        configure = StripToolConfigure(cf)
        self.parameters = Parameter.create(
            name="params", type="group", children=configure.parse())
        self.window.parameterPane.setParameters(self.parameters, showTop=False)

        # Model to be used
        self.model = DataReceiver()

        # Build GUI elements
        self.warning = WarningDialog(None)
        self.pvedit_dialog = PvEditDialog()

    def create_controller(self, cfgtext=None):
        if not cfgtext:
            cfgtext = self.DEFAULT_CFG
        cfg = ConfigParser()
        cfg.read_string(cfgtext)
        self.cfg = cfg
        
        self.striptool_controller = StripToolController(
            self.window,  self.model, self.pvedit_dialog, self.warning, self.parameters, self.cfg)
        
    def test_pvlist(self):
        raw = """
[STRIPTOOL]
Chan1.PV = pva://Bar:Baz
Chan2.PV = pva://Bar:Baz2
Chan1.Color = #000000
Chan2.Color = #FFFFFF
        """

        self.create_controller(raw)
        
        expected = {
            'Bar:Baz' : PvConfig('Bar:Baz', '#000000', 'pva'),
            'Bar:Baz2' : PvConfig('Bar:Baz2', '#FFFFFF', 'pva'),
            }
        pvdict = self.striptool_controller.pvdict
        self.assertEqual(len(expected), len(pvdict))
        for k, v in expected.items():
            self.assertTrue(k in pvdict)
            p = pvdict[k].make_pvconfig()
            self.assertEqual(p.pvname, v.pvname)
            self.assertEqual(p.color, v.color)
            self.assertEqual(p.proto, v.proto)


        newpvlist = [ PvConfig('Foo:bar', '#444444', 'ca'), PvConfig('Foo:bar2', '#555555', 'ca') ]
        self.striptool_controller.pv_edit_callback(newpvlist)
        self.assertEqual(len(newpvlist), len(pvdict))
        pvdict = self.striptool_controller.pvdict
        for v in newpvlist:
            self.assertTrue(v.pvname in pvdict)
            p = pvdict[v.pvname].make_pvconfig()
            self.assertEqual(p.pvname, v.pvname)
            self.assertEqual(p.color, v.color)
            self.assertEqual(p.proto, v.proto)
            

    def test_pvedit(self):
        raw = """
[STRIPTOOL]
DefaultProtocol = pva
Chan0.PV = Ch0
Chan1.PV = Ch1
Chan2.PV = Ch2
"""

        pvcount = 3
        self.create_controller(raw)

        self.assertEqual(len(self.striptool_controller.pvdict), pvcount)
        pvlist = []
        for p in self.striptool_controller.pvdict.values():
            cfg = p.make_pvconfig()
            pvlist.append(cfg)

        pvedit = self.striptool_controller._pvedit_dialog
        pvedit._set_pvlist(pvlist)
        pvedit._add_pv('Ch3', '#000000', 'pva')
        pvedit._on_ok()

        self.assertEqual(len(self.striptool_controller.pvdict), pvcount + 1)
        self.assertTrue('Ch3' in self.striptool_controller.pvdict)

    def test_update_status(self):
        """
        Test update statistics
        """
        self.create_controller()
        
        #make sure that update_status runs without errors
        self.striptool_controller.update_status()
        



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import random
from ..model import *

def randcolor():
    return "#"+''.join([random.choice('0123456789ABCDEF') for j in range(6)])


class PvConfig:
    default_colors = ['#FFFF00', '#FF00FF', '#55FF55', '#00FFFF', '#5555FF',
                      '#FFAA00', '#FF5555', '#0000FF', '#AAAA00', '#FFFFFF']
    color_index = 0
    
    def __init__(self, pvname=None, color=None, proto=None):
        self.pvname = pvname
        if color:
            self.color = color
        elif PvConfig.color_index < len(PvConfig.default_colors):
            self.color = PvConfig.default_colors[PvConfig.color_index]
            PvConfig.color_index += 1
        else:
            self.color = randcolor()

        self.set_proto(proto)

    def set_proto(self, pname):
        self.proto = make_protocol(pname) if pname is not None else pname        

        




"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import logging
import pyqtgraph
from  .pvedit_dialog_controller import PvEditDialogController
import random
from PyQt5 import QtGui
from ..model import ConnectionState
import pvaccess as pva
from ..view import *
from .scope_controller_base import ScopeControllerBase
from ..view.scope_display import PlotChannel as ScopePlotChannel
from pyqtgraph.parametertree import Parameter
from .striptool_config import StriptoolConfig
from .pvconfig import PvConfig
from pyqtgraph.Qt import QtCore

class PvScopeItem:
    def __init__(self, props, controller):
        self.pvname = props.pvname
        self.proto = props.proto
        self.channel = ScopePlotChannel(props.pvname, props.color)
        self.parent_controller = controller
        self.connection = controller.model.create_connection(props.pvname, provider=props.proto)        
        self.status =  str(self.connection.state)
        self.hide = False

    def start(self):
        if not self.connection.is_running():
            self.connection.start(status_callback=self.connection_changed_callback,
                                  routine=self.monitor_callback)

    def set_hide(self, val):
        self.hide = val
        if not val:
            self.channel.pvname = self.pvname
        else:
            self.channel.pvname = 'None'
            
    def stop(self):
        self.connection.stop()
        self.parent_controller._win.graphicsWidget.clear_sample_data(self.pvname)

    def update_properties(self, props):
        self.channel.color = props.color
        if self.proto != props.proto:
            self.stop()
            self.proto = props.proto
            self.connection = self.parent_controller.model.create_connection(self.pvname, provider=self.proto)
            self.connection_changed_callback(self.connection.state, None)
            self.start()
            
    def connection_changed_callback(self, state, msg):
        status = str(state)
        if msg:
            status += ': ' + msg
            
        self.status = status

        if not self.connection.is_running():
            self.parent_controller._win.graphicsWidget.clear_sample_data(self.pvname)
            
        self.parent_controller.connection_changed_signal.sig.emit(self.pvname,
                                                    "%s/%s" % (str(self.proto), status))
        
    def monitor_callback(self, data):
        try:
            data =  data['value']
        except:
            logging.getLogger().error('Unable to plot %s. PV is not scalar. Stopping connection', self.pvname)
            self.stop()
            return

        self.parent_controller.data_callback(self.pvname, data)

    def __del__(self):
        self.connection.stop()

    def make_pvconfig(self):
        return PvConfig(self.pvname, self.channel.color, self.proto)
    

class StripToolController(ScopeControllerBase):
    def __init__(self, widget, model, pvedit_widget, warning, parameters,
                 cfg,
                 **kwargs):

        super().__init__(widget, model, parameters, warning, **kwargs)
        st_config = StriptoolConfig(cfg, **kwargs)
        self._pvedit_dialog = PvEditDialogController(pvedit_widget, model,
                                                     default_proto=st_config.default_proto)
        self._win.editPvButton.clicked.connect(self._on_pvedit_click)
        
        # Signal to update status when connection changes
        class FlagSignal(QtCore.QObject):
            sig = QtCore.pyqtSignal(str, str)
            def __init__(self):
                QtCore.QObject.__init__(self)
        self.connection_changed_signal = FlagSignal()
        self.connection_changed_signal.sig.connect(self.update_status_params)
        
        self.default_config(**kwargs)
        #default to showing 60 second of data
        if not self._win.graphicsWidget.max_length:
            self.update_buffer_samples(int(60000 / self.refresh))

        self._win.graphicsWidget.enable_sampling_mode(True)
        self._pvedit_dialog.set_completion_callback(self.pv_edit_callback)
        self._init_pvlist(st_config.pvs.values())

        self.start_plotting()
        
    def _init_pvlist(self, pvconfig):
        self.pvdict = {}
        pvlist = []
        for p in pvconfig:
            if not p.pvname:
                continue
            
            pvlist.append(p)
        self._set_pv_list(pvlist)
        
    def _on_pvedit_click(self):
        pvlist = []
        for p in self.pvdict.values():
            cfg = p.make_pvconfig()
            pvlist.append(cfg)
            
        self._pvedit_dialog.show(pvlist)

    def pv_edit_callback(self, newpvlist):
        self._set_pv_list(newpvlist)

    def _set_pv_list(self, pvlist):
        oldpvnames_set = set(self.pvdict.keys())
        newpvnames_set = set([p.pvname for p in pvlist])

        newpvs = newpvnames_set - oldpvnames_set
        removed_pvs = oldpvnames_set - newpvnames_set
        
        #remove pvs
        for p in removed_pvs:
            item = self.pvdict[p]
            del self.pvdict[p]
            
        #new pvs        
        for p in pvlist:
            pvname = p.pvname
            if pvname not in self.pvdict:
                si =  PvScopeItem(p, self)
                self.pvdict[pvname] = si
            else:
                si = self.pvdict[pvname]
                si.update_properties(p)
        
        #setup channel config
        self._setup_channel_config()
        
        #create channel list
        self._setup_plot()
        
        for si in self.pvdict.values():
            si.start()
            
    def update_status_params(self, pvname, status_str):
        self.update_channel_param(pvname, "Proto/status", status_str)

    def _setup_plot(self):
        channels = [ si.channel for si in self.pvdict.values() ]
        self._win.graphicsWidget.setup_plot(channels)
        
    def _setup_channel_config(self):
        channel = []
        for si in self.pvdict.values():
            channel.append(
                { "name" : si.pvname,
                  "type": "group",
                  "children": [
                      {"name": "Proto/status", "type": "str",
                       "value": "%s/%s" % (str(si.proto), si.status),
                       "readonly":True},
                      {"name": "Hide", "type": "bool", "value":si.hide},
                      {
                          "name": "Color",
                          "type": "color",
                          "value": si.channel.color
                      },
                      {"name": "DC offset", "type": "float", "value": si.channel.dc_offset},
                      {"name": "Axis location", "type": "list", "limits": {
                          "Left" : "left",
                          "Right" : "right",
                      }, "value" : si.channel.axis_location},
                  ]
                })
        parameters = Parameter.create(name="params", type="group", children=channel)
        parameters.sigTreeStateChanged.connect(self.channel_param_changed)
        self.chan_parameters = parameters
        self._win.channelParamPane.setParameters(parameters, showTop=False)

    def update_channel_param(self, pvname, pname, val):
        param = self.chan_parameters.child(pvname).child(pname)
        param.setValue(val)
        
    def data_callback(self, pvname, data):
        def generator():
            d = {pvname: data}
            
            for k, v in d.items():
                yield k, v
            
        self._win.graphicsWidget.data_process(generator)
        
    def channel_param_changed(self, params, changes):
        for param, change, data, in changes:
            if change != "value":
                continue
            
            child = self.chan_parameters.childPath(param)
            if not child:
                continue
            
            pvname = child[0]
            si = self.pvdict[pvname]
            
            pname = param.name()
            if pname == 'Hide':
                si.set_hide(data)
            elif pname == 'Color':
                c = data.getRgb()
                si.channel.color = '#%02x%02x%02x' % (c[0], c[1], c[2])
            elif pname == 'DC offset':
                si.channel.dc_offset = data
            elif pname == 'Axis location':
                si.channel.axis_location = data

        self._setup_plot()

    def set_sampling_mode(self, val):
        self._win.graphicsWidget.enable_sampling_mode(val)
        
    def parameter_change(self, params, changes):
        """

        :param params:
        :param changes:
        :return:
        """
        for param, change, data in changes:
            if change == "value":
                path = self.parameters.childPath(param)
                if path is not None:
                    childName = '.'.join(path)
                else:
                    childName = param.name()

                if childName == "Acquisition.Sample Mode":
                    self.set_sampling_mode(data)
                    
        super().parameter_change(params, changes)



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

from .pvconfig import PvConfig
import re
from ..model import *


def parse_pv(val):
    pvname = None
    if "://" in val:
        proto, pvname = val.split('://')
        return pvname, proto
    else:
        pvname = val
        return pvname, None
    
class StriptoolConfig():
    def __init__(self, cfg, **kwargs):
        self.default_proto = None
        self.pvs = {}
        
        cfg = cfg['STRIPTOOL']
        if not cfg:
            return
        
        self.default_proto = make_protocol(cfg.get('DefaultProtocol', 'ca'))

        cfgpvs = {}
        for k, v in cfg.items():
            if bool(re.match('chan', k, re.I)):
                ch, param = k.split('.')
                chcfg = cfgpvs.get(ch, PvConfig())

                if param == 'pv':
                    pvname, proto = parse_pv(v)
                    chcfg.pvname = pvname
                    if proto is None:
                        proto = self.default_proto
                    chcfg.set_proto(proto)
                    
                elif param == 'color':
                    chcfg.color = str(v)
                    
                cfgpvs[ch] = chcfg

        for p in cfgpvs.values():
            self.pvs[p.pvname] = p
            
        pvs = kwargs.get('pv')
        if pvs:
            for pv in pvs.values():
                chcfg = None
                pvname, proto = parse_pv(pv)
                if pvname in self.pvs:
                    chcfg = self.pvs[pvname]
                else:
                    chcfg = PvConfig()
                    chcfg.set_proto(self.default_proto)
                    
                chcfg.pvname = pvname
                if proto:
                    chcfg.set_proto(proto)
                self.pvs[pvname] = chcfg                
                



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2021 UChicago Argonne LLC
 as operator of Argonne National Laboratory

Unit tests for ImageProfileWidget

@author: Matic Pogacnik <mpogacnik@anl.gov>
"""
import sys
import os
import unittest
import unittest.mock as mock
import numpy as np
import pyqtgraph as pg

from c2dataviewer.imagev import ImageWindow, ImageSettingsDialog, WarningDialog
from c2dataviewer.control.imagecontroller import ImageController
from c2dataviewer.view.image_definitions import COLOR_MODE_MONO, COLOR_MODE_RGB2, transcode_image
from c2dataviewer.view.image_profile_display import ImageProfileWidget
from .helper import create_image
from .helper import setup_qt_app

os.environ["QT_QPA_PLATFORM"] = "offscreen"


class TestImageProfileWidget(unittest.TestCase):

    
    def setUp(self):
        """
        Build the environment for each unit test case.
        This method is called before each test.
        We need the whole application to be able to test this, as ImageProfileWidget also handle the layout.

        :return:
        """
        # Create Qt application
        setup_qt_app()
        
        # Create ImageWindow and het the imageWidget instance
        self.window = ImageWindow()
        self.imageWidget = self.window.imageWidget

        # Build GUI elements
        w = ImageWindow(None)
        datasource = mock.Mock()
        w.imageWidget.set_datasource(datasource)
        image_settings_dialog = ImageSettingsDialog(None)
        warning = WarningDialog(None)

        self.ic = ImageController(w, IMAGE_SETTINGS_DIALOG=image_settings_dialog, WARNING=warning,
                PV={'test', 'test'}, timer=pg.Qt.QtCore.QTimer(), data=None)

        
        # Get ImageProfileWidget
        self.w = self.ic._win.imageWidget.image_profile_widget


    def tearDown(self):
        """
        Tear down the environment after each test case.
        This mentod is called after each test.

        :return:
        """

    def test_calculate_profiles(self):
        """
        Test if profile calculation is correct.

        :return:
        """
        # Build test image
        arrayValue = [
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            ]

        # Test with the mono image
        data = create_image(1, arrayValue, data_type='ubyteValue', nx=10, ny=10, color_mode=COLOR_MODE_MONO)
        data = transcode_image(data['value'][0]['ubyteValue'], COLOR_MODE_MONO, 10, 10, None)
        self.w._display_profiles = True
        self.w.set_image_data(data, COLOR_MODE_MONO)
        x_profile, y_profile = self.w._calculate_profiles()

        np.testing.assert_equal(np.array([68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3],
                                        dtype=np.double),
                                x_profile
                        )
        np.testing.assert_equal(np.array([255.,   0.,  77.,  54.,  23.,  76.,  34.,  65.,  34.,  65.],
                                        dtype=np.double),
                                y_profile
                        )

        # Test with the color image
        data = create_image(3, arrayValue*3, data_type='ubyteValue', nx=10, ny=10, nz=3, color_mode=COLOR_MODE_RGB2)
        data = transcode_image(data['value'][0]['ubyteValue'], COLOR_MODE_RGB2, 10, 10, 3)
        self.w.set_image_data(data, COLOR_MODE_RGB2)
        x_profile, y_profile = self.w._calculate_profiles()

        np.testing.assert_equal(np.array([  [68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3],
                                            [68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3],
                                            [68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3, 68.3],],
                                        dtype=np.double),
                                x_profile
                        )
        np.testing.assert_equal(np.array([  [255.,   0.,  77.,  54.,  23.,  76.,  34.,  65.,  34.,  65.],
                                            [255.,   0.,  77.,  54.,  23.,  76.,  34.,  65.,  34.,  65.],
                                            [255.,   0.,  77.,  54.,  23.,  76.,  34.,  65.,  34.,  65.],],
                                        dtype=np.double),
                                y_profile
                        )




# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities

@author: Guobao Shen <gshen@anl.gov>
"""


import os
import sys
import pkg_resources
import argparse
from configparser import ConfigParser
import enum
import logging
import faulthandler

class AppType(enum.Enum):
    SCOPE = "scope"
    IMAGE = "image"
    STRIPTOOL = "striptool"

    def __str__(self):
        return self.value

def qxl_module_loaded():
    fn = "/proc/modules"
    if not os.path.isfile(fn):
        return False
    with open(fn, "r") as f:
        modules = f.readlines()
    return any([m for m in modules if m.startswith("qxl")])

# Suppress libGL error about swrast when running
# on a VM with paravirtual graphics driver (qxl).
# Run before any qt code runs
if qxl_module_loaded():
    os.environ['LIBGL_ALWAYS_INDIRECT'] = 'true'

# Limit numpy to a single thread
# Must run before numpy import
os.environ['MKL_NUM_THREADS'] = '1'
os.environ['NUMEXPR_NUM_THREADS'] = '1'
os.environ['OMP_NUM_THREADS'] = '1'

# Get the version number used by setup.py at package build time
try:
    __version__ = pkg_resources.get_distribution('C2DataViewer').version
except pkg_resources.DistributionNotFound:
    __version__ = 'Unknown'


def load_config(N=None):
    """
    Load default configuration from configuration file

    :return:
    """
    import os.path
    pkgcfg = os.path.join( os.path.dirname(__file__), 'c2dv.cfg' )
    cfgfile = ['/etc/c2dv.cfg', os.path.expanduser('~/.c2dvrc'), 'c2dv.cfg', pkgcfg]
    # if N is not None:
    #     cfgfile.append(N)

    noconfig = True
    for cfg in cfgfile:
        if os.path.isfile(cfg):
            noconfig=False
            break

    if noconfig:
        raise RuntimeError("No configuration file found.")
    cf = ConfigParser()
    if N is not None:
        if not os.path.isfile(N):
            raise RuntimeError("%s not found." % (N))
        cf.read(N)
    else:
        cf.read(cfgfile)
    return cf


def pvmaps(pvs, alias=None):
    """

    :param pvs: EPICS PV name(s)
    :param alias: EPICS PV alias name(s)
    :return:
    """
    pvmap = {}
    # labels = []
    if alias is not None:
        alias = alias.split(",")
        for idx, pv in enumerate(pvs.split(",")):
            pv = pv.strip()
            if pv != "":
                if idx < len(alias) and alias[idx].strip() != "":
                    pvmap[alias[idx].strip()] = pv.strip()
                else:
                    pvmap[pv.strip()] = pv.strip()
    else:
        for pv in pvs.split(","):
            pvmap[pv.strip()] = pv.strip()
    return pvmap


def main():
    """

    :return:
    """    
    parser = argparse.ArgumentParser(
            description='C2DataViewer for EPICS7 PVStructure Data display via pvAccess')

    parser.add_argument('--app', type=AppType,
                        help='Application name', choices=list(AppType))
    # TODO: add CLI interface later to over write configuration
    parser.add_argument('--config', type=str,
                        help='configuration file')
    parser.add_argument('--pv', type=str,
                        help='EPICS PV names. Support multiple PVs which are comma separated.'
                             ' e.g. PV1,PV2,PV3')
    parser.add_argument('--alias', type=str,
                        help='EPICS PV alias names. Support multiple alias name which are comma separated.'
                             ' e.g. name1,name2,name3')
    parser.add_argument('--fields', help='Comma-separate list of PV fields to display for scope application')
    parser.add_argument('--connect-on-start', action='store_true',
                        help='Connect PV on start up for scope application')
    parser.add_argument('--arrayid', type=str,
                        help='EPICS PV field name for ID for scope application, '
                             'which is used by scope to count losing arrays.'
                             ' e.g. ArrayId')
    parser.add_argument('--xaxes', type=str,
                        help='EPICS PV field name to be used for x axes for scope application.'
                             ' e.g. Time')
    parser.add_argument('--max', type=float, default=None,
                        help='Filter out values greater than MAX')
    parser.add_argument('--min', type=float, default=None,
                        help='Filter out values less than MIN')
    parser.add_argument('--trigger', type=str, default=None,
                        help='Trigger PV for scope. Format: [proto://]pv_name. '
                             'Protocol is optional, and needs to be either "ca" or "pva" when specified, '
                             'by default, uses EPICS7 pvAccess.'
                             ' e.g. ca://MY:TEST:TRIG:PV or MY:TEST:TRIG:PV')
    parser.add_argument(
        "--debug",
        action='store_true',
        help="Turn on debug messages. Can also turn this on by setting environment variable C2DV_LOG_LEVEL=DEBUG"
    )

    args = parser.parse_args()

    # Dump backtrace to stderr on SEGFAULT
    faulthandler.enable()
    
    level = logging.CRITICAL

    if os.getenv('C2DV_LOG_LEVEL'):
        level = os.getenv('C2DV_LOG_LEVEL').upper()
        
    if args.debug:
        level = logging.DEBUG
        
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(asctime)s %(levelname)-6s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    
    cfg = load_config(args.config)

    pv_map = None
    if args.pv is not None:
        pv_map = pvmaps(args.pv, args.alias)

    app = args.app
    if args.app is None:
        app = AppType(cfg["DEFAULT"]["APP"].lower())
        
    if app == AppType.IMAGE:
        from c2dataviewer import imagev
        # application from CLI overwrite the one in configuration file
        section = cfg[cfg["DEFAULT"]["APP"]]["SECTION"]
        if pv_map is None:
            # use pv information from configuration file
            try:
                pvnames = cfg[section]["PV"] or None
            except KeyError:
                pvnames = None
            try:
                alias = cfg[section]["ALIAS"] or None
            except KeyError:
                alias = None
            if pvnames is not None:
                pv_map = pvmaps(pvnames, alias)
        try:
            scale = float(cfg[section]["SCALE"])
        except KeyError:
            scale = 1.0
        try:
            noAGC = True if cfg[section]["AUTOGAIN"] == 'DISABLED' else False
        except KeyError:
            noAGC = False
        imagev(pv_map, scale, noAGC)
    elif app == AppType.SCOPE:
        from c2dataviewer import scope
        scope(cfg, pv=pv_map, arrayid=args.arrayid, xaxes=args.xaxes,
              max=args.max, min=args.min, trigger=args.trigger,
              connect_on_start=args.connect_on_start,
              fields=args.fields)
    elif app == AppType.STRIPTOOL:
        from c2dataviewer import striptool
        striptool(cfg, pv=pv_map)
    else:
        raise RuntimeError("Unknown application ({0})".format(args.app))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities using pvaPy as pvAccess binding

@author: Guobao Shen <gshen@anl.gov>
"""
import time
from threading import Lock
from collections import deque
from statistics import mean
import enum

import pvaccess as pva
from pvaccess import PvaException

def make_protocol(proto):
    """
    Returns the pvapy enum for the given protocol string
    
    proto:  "ca" or "pva"
    """
    if isinstance(proto, pva.ProviderType):
        return proto
    
    proto = proto.lower()
    if proto == 'ca':
        return pva.ProviderType.CA
    elif proto == 'pva':
        return pva.ProviderType.PVA
    else:
        raise Exception('Invalid protocol: ' + proto)

def parse_pvname(pvname, default_proto):
    """
    Parses the pvname string provided from user.  Returns
    (PV, protocol)

    pvname:  string in format proto://PV or PV
    default_proto:  Default protocol if one is not given
    """
    proto = default_proto
    if "://" in pvname:
        proto, pvname = pvname.split('://')
        proto = make_protocol(proto.lower().strip())

    return pvname, proto

class PollStrategy:
    """
    Implementation of strategy interface used in Channel
    
    Collects data from the PV Channel by periodically polling
    the channel.
    """
    
    def __init__(self, context, timer):
        """
        context: the Channel object
        timer: QT Timer object 
        """
        self.ctx = context
        self.timer = timer
        self.timer.timeout.connect(self.poll)
        
    def _data_callback(self, data):
        #
        # Callback called with returned data from the PV channel
        #
        # data: data from PV channel
        #
        # Set state to CONNECTING when data returned
        self.ctx.set_state(ConnectionState.CONNECTING)
        self.ctx.data_callback_wrapper(data)

    def _err_callback(self, msg):
        #
        # Callback called if there was an error attempting to get data
        # Notifies channel that error occured
        #
        # msg: error message
        #
        self.ctx.notify_error(msg)
        
    def poll(self):
        """
        Asynchronously calls pvget on PV channel.  
        Called by the timer object.
        Returned data and any errors are passed in via callbacks
        """
        try:
            self.ctx.channel.asyncGet(self._data_callback, self._err_callback, '')
        except pva.PvaException:
            #error because too many asyncGet calls at once.  Ignore
            pass
        
    def start(self):
        """
        Starts timer to poll data. 
        """
        self.timer.start(int(1000/self.ctx.rate))

    def stop(self):
        """
        Stops polling
        """
        self.timer.stop()
    
class MonitorStrategy:

    """
    Implementation of strategy interface used in Channel

    Uses pvmonitor to get data from PV channel
    """
    def __init__(self, context):
        """
        context:  Channel object
        """
        self.ctx = context
        
    def _data_callback(self, data):
        #
        # Data callback called by pvmonitor
        #
        # data:  data from PV channel
        #
        self.ctx.data_callback_wrapper(data)

    def _connection_callback(self, is_connected):
        #
        # Callback called by pvmonitor when there is a connection change
        # Changes the Channel state or notify error
        #
        #  is_connected:  True if connected to PV channel.  Otherwise False
        #
        if is_connected:
            self.ctx.set_state(ConnectionState.CONNECTED)
        elif self.ctx.is_running():
            self.ctx.notify_error()            
        
    def start(self):
        """
        Starts the PV monitor
        """
        try:
            self.ctx.channel.setConnectionCallback(self._connection_callback)
            request = 'field()'
            if self.ctx.data_source.server_queue_size:
                # This establishes image queue on the IOC side, which helps reduce
                # dropped frames at high rates.
                request = f'record[queueSize={self.ctx.data_source.server_queue_size}]field()'
            self.ctx.channel.monitor(self._data_callback, request)
        except PvaException as e:
            self.ctx.notify_error(str(e))
        
    def stop(self):
        """
        Stops the PV monitor
        """
        try:
            self.ctx.channel.setConnectionCallback(None)
            self.ctx.channel.stopMonitor()
            self.ctx.channel.unsubscribe('monitorCallback')
        except PvaException:
            pass


class ConnectionState(enum.Enum):
    """
    Connection State of the PV channel
    """
    CONNECTED = 1
    CONNECTING = 2
    DISCONNECTING = 3
    DISCONNECTED = 4
    FAILED_TO_CONNECT = 5

    def __str__(self):
        string_lookup = {
            1 : 'Connected',
            2 : 'Connecting',
            3 : 'Disconnecting',
            4 : 'Disconnected',
            5 : 'Failed to connect'
        }

        return string_lookup[int(self.value)]

class Channel:
    """
    PV Channel object.  Manages getting the PV data and monitoring channel connection
    status
    """
    def __init__(self, name, data_source, timer=None, provider=pva.PVA, status_callback=None):
        """
        name : PV name
        data_source: data source class that owns this object
        timer: Timer to use if polling data
        provider: PV protocol
        status_callback:  Callback called if connection state changed
        """
        self.channel = pva.Channel(name, provider)
        self.data_source = data_source
        self.provider = provider
        self.name = name
        self.rate = None
        self.data_callback = None
        self.data = None
        self.monitor_strategy = MonitorStrategy(self)
        self.poll_strategy = PollStrategy(self, timer) if timer else None
        self.strategy = None
        self.rate = None
        self.status_callback = status_callback
        self.state = ConnectionState.DISCONNECTED

    def data_callback_wrapper(self, data):
        """
        Data callback called when received data from the PV channel
        
        data: PV data
        """
        if not self.is_running():
            return
        
        self.set_state(ConnectionState.CONNECTED)
        if self.data_callback:
            self.data_callback(data)
        else:
            self.data = data.get()

    def notify_error(self, msg=None):
        """
        Notify that unable to connect to PV channel
        """
        #Only update status for the current PV that is running to prevent status overrides during error callbacks
        if self.is_running():
            self.set_state(ConnectionState.FAILED_TO_CONNECT, msg)
        
    def start(self, routine=None, rate=None, status_callback=None):
        """
        Start monitoring data
        
        routine: PV data callback
        rate: Rate to poll data.  If set, will poll to get data.  Otherwise
             will use a PV monitor
        status_callback: status callback.  Called whenever there is a state change.
        """
        self.data_callback = routine
        self.rate = rate
        self.strategy = self.poll_strategy if self.rate else self.monitor_strategy
        if not self.strategy:
            raise Exception("Can't poll data unless DataSource timer is configured")
        if status_callback:
            self.status_callback = status_callback
        self.set_state(ConnectionState.CONNECTING)
        self.strategy.start()

    def set_state(self, state, msg=None):
        """
        Set connection state.  

        state: ConnectionState object
        msg: (str) optional message
        """
        if state != self.state:
            self.state = state
            if self.status_callback:
                self.status_callback(str(state), msg)
                
    def stop(self):
        """
        Stop monitoring data
        """
        self.set_state(ConnectionState.DISCONNECTING)
        if self.strategy:
            self.strategy.stop()
        self.set_state(ConnectionState.DISCONNECTED)

    def is_running(self):
        """
        Returns True if Channel is running
        """
        return self.state in [ConnectionState.CONNECTED, ConnectionState.CONNECTING]
    
    def get(self):
        """
        Get data from PV channel.  blocking call.  Returns data if successful.
        """
        try:
            return self.channel.get('')
        except PvaException as e:
            self.notify_error(str(e))
            raise

    def async_get(self, data_callback, error_callback):
        """
        Asynchronously get data from PV channel. Nonblocking call.
        
        data_callback: callback called when received data
        error_callback: callback called if error occured
        """
        try:
            self.channel.asyncGet(data_callback, error_callback, '')
        except pva.PvaException:
            #error because too many asyncGet calls at once.  Ignore
            pass
        
class DataSource:
    def __init__(self, timer_factory=None, default=None):
        """

        :param default:
        """
        # local cache of latest data
        self.data = None
        # EPICS7 channel
        self.channel = None
        self.channel_cache = {}
        self.fps = None
        self.server_queue_size = 0
        self.timer_factory = timer_factory if timer_factory else lambda: None
            
        self.data_callback = None
        self.status_callback = None
        
        # Default PV name
        self.device = None
        if default is not None:
            if type(default) in [list, tuple]:
                self.__init_connection(default[0])
            elif type(default) == dict:
                self.__init_connection(list(default.values())[0])
            elif type(default) == str:
                self.__init_connection(default)
            else:
                raise RuntimeError("Unknown data type of default parameter: ", default)

        # trigger support from external EPICS3/7 channel
        self.trigger = None
        self.trigger_chan = None

    def create_connection(self, name, provider):
        if name in self.channel_cache and self.channel_cache[name].provider == provider:
            return self.channel_cache[name]
        else:
            chan = Channel(name, self, self.timer_factory(), provider=provider)
            self.channel_cache[name] = chan
            return chan
    
    def __init_connection(self, name):
        """
        Create initial channel connection with given PV name

        :param name: EPICS7 PV name
        :return:
        """

        name, proto = parse_pvname(name, pva.ProviderType.PVA)
        self.device = name
        
        self.channel = Channel(self.device, self, self.timer_factory(), proto)
        self.channel_cache[name] = self.channel
            
    def get(self):
        """
        Get data from current PV channel
        :return:
        """
        if self.channel is None:
            return None

        if not self.channel.status_callback:
            self.channel.status_callback = self.status_callback
        
        return self.channel.get()

    def async_get(self, success_callback=None, error_callback=None):
        """
        Get data from current PV channel
        :return:
        """
        if self.channel is None:
            return

        if not self.channel.status_callback:
            self.channel.status_callback = self.status_callback

        if not success_callback:
            success_callback = lambda x: None

        if not error_callback:
            error_callback = lambda x: None
            
        self.channel.async_get(success_callback, error_callback)

    def update_framerate(self, fps):
        self.fps = fps
        
    def update_server_queue_size(self, sqs):
        self.server_queue_size = sqs

    def update_device(self, name, restart=False, test_connection=True):
        """
        Update device, EPICS PV name, and test its connectivity

        :param name: device name
        :param restart: flag to restart or not
        :param test_connection: flag to test channel connectivity or not
        :return:
        :raise PvaException: raise pvaccess exception when channel cannot be connected.
        """
        if self.channel is not None:
            self.stop()

        if name != "":
            chan = None
            if name in self.channel_cache:
                chan = self.channel_cache[name]
            else:
                name, proto = parse_pvname(name, pva.ProviderType.PVA)
                chan = Channel(name, self, self.timer_factory(), provider=proto, status_callback=self.status_callback)
                self.channel_cache[name] = chan
                
            self.channel = chan
            self.device = name

            # test channel connectivity
            if test_connection:
                chan.get()

            if restart:
                self.start()
        else:
            self.channel = None

    def start(self, routine=None, status_callback=None):
        """
        Start a EPICS7 monitor for current device.

        :return:
        :raise PvaException: raise pvaccess exception when channel cannot be connected.
        """
        if self.channel is None:
            return

        if routine:
            self.data_callback = routine

        if status_callback:
            self.status_callback = status_callback
            
        self.channel.start(routine=self.data_callback, rate=self.fps, status_callback=self.status_callback)

    def stop(self):
        """
        Stop monitor EPICS7 channel for current device.

        :return:
        :raise PvaException: raise pvaccess exception when channel fails to disconnect.
        """
        if self.channel is None:
            return

        self.channel.stop()

    def update_trigger(self, name, proto='ca'):
        """
        Update trigger PV name.  If failed to connect, trigger is set to None

        :param name: PV name for external trigger
        :param proto: protocol, should be 'pva' or 'ca'. 'ca' by default
        :return: trigger fields
        """
        if self.trigger_chan is not None:
            self.stop_trigger()

        self.trigger = None
        
        if name == self.trigger:
            # nothing change
            return
        
        if proto.lower() == "ca":
            # use channel access protocol, which is default
            trigger_chan = pva.Channel(name, pva.CA)
        elif proto.lower() == "pva":
            # use pvAccess protocol
            trigger_chan = pva.Channel(name)
        else:
            raise RuntimeError("Unknown EPICS communication protocol {}".format(proto))

        # test connectivity, which might cause for example timeout exception
        pv = trigger_chan.get("")

        # update trigger name & channel
        self.trigger = name
        self.trigger_proto = proto.lower()
        self.trigger_chan = trigger_chan
        pv_structure = pv.getStructureDict()
        keys =[ k for k,v in pv_structure.items() ]
        
        return keys

    def start_trigger(self, routine):
        """

        :return:
        """
        try:
            self.trigger_chan.subscribe('triggerMonitorCallback', routine)
            req = ''
            if self.trigger_proto == 'ca':
                #need to explicitly ask for timeStamp 
                req = 'field(timeStamp,value)'
            self.trigger_chan.startMonitor(req)
        except PvaException:
            raise RuntimeError("Cannot connect to PV {}".format(self.trigger))

    def stop_trigger(self):
        """

        :return:
        """
        if not self.trigger_chan:
            return
        
        try:
            self.trigger_chan.stopMonitor()
            self.trigger_chan.unsubscribe('triggerMonitorCallback')
        except PvaException:
            # raise RuntimeError("Fail to disconnect")
            pass



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import unittest
from configparser import ConfigParser
from c2dataviewer.view.scopeconfig import Configure


class TestScopeConfig(unittest.TestCase):
    def get_display_val(self, data, val):
        for child in data['children']:
            if child['name'] == val:
                ret_val = child['value']
                return ret_val
            
    def test_autoscale(self):
        #Does autoscale setting in app specific section take precedence
        raw1 = """
        [SCOPE]
        DefaultProtocol = ca
        AUTOSCALE=False

        [DISPLAY]
        AUTOSCALE=True
        AVERAGE=1
        HISTOGRAM=False
        N_BIN=100
        REFRESH=100
        """
        parser = ConfigParser()
        parser.read_string(raw1)
        configure = Configure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)

        self.assertFalse(self.get_display_val(data=display, val='Autoscale'))

        #When autoscale setting absent in app specific section, but present in DISPLAY
        raw2 = """
        [SCOPE]
        DefaultProtocol = ca

        [DISPLAY]
        AUTOSCALE=True
        AVERAGE=1
        HISTOGRAM=False
        N_BIN=100
        REFRESH=100
        """
        parser = ConfigParser()
        parser.read_string(raw2)
        configure = Configure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)
        
        self.assertTrue(self.get_display_val(data=display, val='Autoscale'))

        #When autoscale setting absent in both app specific and in DISPLAY sections,
        #is default selected
        raw3 = """
        [SCOPE]
        DefaultProtocol = ca

        [DISPLAY]
        AVERAGE=1
        HISTOGRAM=False
        N_BIN=100
        REFRESH=100
        """
        parser = ConfigParser()
        parser.read_string(raw3)
        configure = Configure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)

        self.assertFalse(self.get_display_val(data=display, val='Autoscale'))



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2021 UChicago Argonne LLC
 as operator of Argonne National Laboratory

Unit tests for image_definitions.py

@author: Matic Pogacnik <mpogacnik@anl.gov>
"""
import os
import unittest
import numpy as np

from c2dataviewer.view.image_definitions import *
from .helper import create_image

os.environ["QT_QPA_PLATFORM"] = "offscreen"

class TestImageDefinitions(unittest.TestCase):

    def setUp(self):
        """
        Build the environment for each unit test case.
        This method is called before each test.

        :return:
        """
        pass


    def tearDown(self):
        """
        Tear down the environment after each test case.
        This mentod is called after each test.

        :return:
        """
        pass

    def test_transcode_image(self):

        x = 10
        y = 10

        arrayValue = [
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            ]


        # Mono image
        data = create_image(1, arrayValue, data_type='ubyteValue', nx=x, ny=y, color_mode=COLOR_MODE_MONO)
        img = transcode_image(data['value'][0]['ubyteValue'], COLOR_MODE_MONO, 10, 10, None)
        np.testing.assert_equal(np.array([[255,   0,  77,  54,  23,  76,  34,  65,  34,  65],
                                        [255,   0,  77,  54,  23,  76,  34,  65,  34,  65],
                                        [255,   0,  77,  54,  23,  76,  34,  65,  34,  65],
                                        [255,   0,  77,  54,  23,  76,  34,  65,  34,  65],
                                        [255,   0,  77,  54,  23,  76,  34,  65,  34,  65],
                                        [255,   0,  77,  54,  23,  76,  34,  65,  34,  65],
                                        [255,   0,  77,  54,  23,  76,  34,  65,  34,  65],
                                        [255,   0,  77,  54,  23,  76,  34,  65,  34,  65],
                                        [255,   0,  77,  54,  23,  76,  34,  65,  34,  65],
                                        [255,   0,  77,  54,  23,  76,  34,  65,  34,  65],], dtype=np.uint8),
                                img
                                )

        # RGB1 image
        data = create_image(2, arrayValue*3, data_type='ubyteValue', nx=x, ny=y, nz=3, color_mode=COLOR_MODE_RGB1)
        img = transcode_image(data['value'][0]['ubyteValue'], COLOR_MODE_RGB1, 10, 10, 3)
        np.testing.assert_equal( np.array([[[255,   0,  77],
                                            [ 54,  23,  76],
                                            [ 34,  65,  34],
                                            [ 65, 255,   0],
                                            [ 77,  54,  23],
                                            [ 76,  34,  65],
                                            [ 34,  65, 255],
                                            [  0,  77,  54],
                                            [ 23,  76,  34],
                                            [ 65,  34,  65],] for i in range(x)], dtype=np.uint8),
                                img
                                )

        # RGB2 image
        data = create_image(3, arrayValue*3, data_type='ubyteValue', nx=x, ny=y, nz=3, color_mode=COLOR_MODE_RGB2)
        img = transcode_image(data['value'][0]['ubyteValue'], COLOR_MODE_RGB2, 10, 10, 3)
        np.testing.assert_equal( np.array([[[255, 255, 255,],
                                            [  0,   0,   0,],
                                            [ 77,  77,  77,],
                                            [ 54,  54,  54,],
                                            [ 23,  23,  23,],
                                            [ 76,  76,  76,],
                                            [ 34,  34,  34,],
                                            [ 65,  65,  65,],
                                            [ 34,  34,  34,],
                                            [ 65,  65,  65,],] for i in range(x)], dtype=np.uint8),
                                img
                                )



        # RGB3 image
        data = create_image(4, arrayValue*3, data_type='ubyteValue', nx=x, ny=y, nz=3, color_mode=COLOR_MODE_RGB3)
        img = transcode_image(data['value'][0]['ubyteValue'], COLOR_MODE_RGB3, 10, 10, 3)
        np.testing.assert_equal( np.array([[[255, 255, 255,],
                                            [  0,   0,   0,],
                                            [ 77,  77,  77,],
                                            [ 54,  54,  54,],
                                            [ 23,  23,  23,],
                                            [ 76,  76,  76,],
                                            [ 34,  34,  34,],
                                            [ 65,  65,  65,],
                                            [ 34,  34,  34,],
                                            [ 65,  65,  65,],] for i in range(x)], dtype=np.uint8),
                                img
                                )



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import subprocess
import re
from setuptools import setup, find_packages





def get_version():
    """
    Get version from the git repository in the following format (This should be compatible with PEP440):
        - If the current commit has a tag and no uncommitted changes function will return only tag as a string.
        - If the current commit doesn't have a tag or there are uncommitted changes, version string will look like:
            <last tag in the current development line>_<offset of the last commit from the last tag>_<last commit>[<_dirty if there are uncommitted changes>]
    """
    try:
        version_git = subprocess.check_output(["git", "describe", "--tags", "--long", "--dirty"]).rstrip()
        match = re.search(r'(.*)-(\d+)-g([0-9,a-f]{7})-?(dirty)?', version_git.decode())
        version = match[1]
        tagOffset = int(match[2])
        hash = match[3]
        dirty = match[4]

        if tagOffset == 0 and not dirty:
            VERSION = version
        else:
            VERSION = f"{version}+{tagOffset}.{hash}{'.' if dirty else ''}{dirty if dirty else ''}"

    except:
        print(f"Error: Can not determinate version from git. Using default value of '0.0.1'")
        VERSION = "0.0.1"

    return VERSION

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup (
    name = "c2dataviewer",
    version = get_version(),
    author = "G. Shen",
    author_email = "gshen@anl.gov",
    description = "Python based data viewer for next generation of APS control system (C2)",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/epics-extensions/c2dataviewer",
    license = "EPICS",
    packages = find_packages(),
    package_data = {'c2dataviewer': ['c2dv.cfg', '**/*.ui']},
    include_package_data = True,
    install_requires=[
        'psutil',
        'pvapy',
        'pyqtgraph>=0.12',
        'PyQt5',
    ],
    extras_require = {
        'blosc-compression' :  ['blosc'],
        'lz4-compression'   :  ['lz4'],
        'bslz4-compression' :  ['bitshuffle'],
        'all'               :  ['blosc', 'lz4', 'bitshuffle'],
    },
    entry_points = {'console_scripts': ['c2dv=c2dataviewer.c2dv:main']},
)
